// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

use {
    crate::gethostbyname::HostEnt,
    crate::nss_bindings::gaih_addrtuple,
    anyhow::{bail, Result},
    libc::{AF_INET, AF_INET6},
    nix::errno::Errno,
    once_cell::sync::Lazy,
    std::ffi::CString,
    std::mem::{align_of, size_of},
    trust_dns_proto::rr::RData,
    trust_dns_resolver::lookup::Lookup,
};

static RUNTIME: Lazy<std::io::Result<tokio::runtime::Runtime>> = Lazy::new(|| {
    // The runtime should remain single-threaded, some
    // programs depend on it (e.g. programs calling unshare())
    let rt = tokio::runtime::Builder::new_current_thread()
        .enable_time()
        .enable_io()
        .build()?;
    Ok(rt)
});

// TODO: error handling codes chosen a bit sloppily
pub unsafe fn records_to_gaih_addr(
    lookup: Lookup,
    mut buf: &mut [u8],
) -> std::io::Result<*mut gaih_addrtuple> {
    const GAIH_ADDRTUPLE_SZ: usize = size_of::<gaih_addrtuple>();

    let mut ret = std::ptr::null_mut();
    let mut prev_link: *mut *mut gaih_addrtuple = std::ptr::null_mut();

    for record in lookup.record_iter() {
        // First add the name to the buffer
        let name = CString::new(record.name().to_utf8())
            .map_err(|e| std::io::Error::new(std::io::ErrorKind::InvalidData, e.to_string()))?;

        let offset = buf.as_ptr().align_offset(align_of::<libc::c_char>());
        let name_src = name.as_bytes_with_nul();
        let name_dest = buf.as_mut_ptr().add(offset);

        let l = name_src.len();
        if buf.len() < offset + l {
            return Err(Errno::ERANGE.into());
        }

        std::ptr::copy_nonoverlapping(name_src.as_ptr(), name_dest, l);
        buf = &mut buf[(offset + l)..];

        // Then add the tuple with the address
        let offset = buf.as_ptr().align_offset(align_of::<gaih_addrtuple>());
        let l = GAIH_ADDRTUPLE_SZ;
        if buf.len() < offset + l {
            return Err(Errno::ERANGE.into());
        }

        let tuple = &mut *(buf.as_mut_ptr().add(offset) as *mut gaih_addrtuple);
        tuple.next = std::ptr::null_mut();
        tuple.name = name_dest as *mut i8;
        tuple.scopeid = 0; // how to set tuple.scopeid ????
        set_if_valid(prev_link, &mut *tuple); // link from previous tuple to this tuple
        prev_link = &mut (*tuple).next;

        match record.data() {
            Some(RData::A(addr)) => {
                tuple.family = AF_INET;
                tuple.addr[0] = std::mem::transmute_copy(&addr.octets());
            }
            Some(RData::AAAA(addr)) => {
                tuple.family = AF_INET6;
                tuple.addr = std::mem::transmute_copy(&addr.octets());
            }
            Some(_) => return Err(Errno::EBADMSG.into()),
            None => return Err(Errno::ENODATA.into()),
        }

        if ret == std::ptr::null_mut() {
            ret = tuple;
        }

        buf = &mut buf[(offset + l)..];
    }

    Ok(ret)
}

pub unsafe fn records_to_hostent(
    lookup: Lookup,
    _hostent: &mut HostEnt,
    mut _buf: &mut [u8],
) -> std::io::Result<()> {

    // char   *h_name       Official name of the host.
    // char  **h_aliases    A pointer to an array of pointers to
    //                      alternative host names, terminated by a
    //                      null pointer.
    // int     h_addrtype   Address type.
    // int     h_length     The length, in bytes, of the address.
    // char  **h_addr_list  A pointer to an array of pointers to network
    //                      addresses (in network byte order) for the host,
    //                      terminated by a null pointer.

    // In C struct hostent:
    //
    // - for the type of queries we perform, we can assume h_aliases
    //   is an empty array.
    // - hostent is limited to just one address type for all addresses
    //   in the list. We pick the type of first result, and only
    //   append addresses of the same type.

    for record in lookup.record_iter() {
        todo!();
    }

    Ok(())
}

pub fn set_if_valid<T>(ptr: *mut T, val: T) {
    if !ptr.is_null() {
        unsafe { *ptr = val };
    }
}

pub fn block_on<F>(f: F) -> Result<F::Output>
where
    F: std::future::Future,
{
    use std::ops::Deref;
    match RUNTIME.deref() {
        Ok(rt) => Ok(rt.block_on(async { f.await })),
        Err(e) => bail!("Unable to start tokio runtime: {}", e),
    }
}
